# Copyright (c) 2018, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

"""Submodule for operating on LVC-format waveform files"""

from .._version import __version__

from .file_io import (read_from_h5,)
